// Copyright 2018-2026 the Deno authors. MIT license.
// @generated by cli/snapshot/build.rs - do not edit.

pub static RESIDUAL_LAZY_JS: &[(&str, &str)] = &[
  ("ext:deno_cache/01_cache.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_cache_01_cache_js.js"))),
  ("ext:deno_canvas/02_surface.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_canvas_02_surface_js.js"))),
  ("ext:deno_cron/01_cron.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_cron_01_cron_ts.js"))),
  ("ext:deno_fetch/20_headers.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_fetch_20_headers_js.js"))),
  ("ext:deno_fetch/21_formdata.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_fetch_21_formdata_js.js"))),
  ("ext:deno_fetch/22_body.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_fetch_22_body_js.js"))),
  ("ext:deno_fetch/23_request.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_fetch_23_request_js.js"))),
  ("ext:deno_fetch/23_response.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_fetch_23_response_js.js"))),
  ("ext:deno_fetch/26_fetch.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_fetch_26_fetch_js.js"))),
  ("ext:deno_fetch/27_eventsource.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_fetch_27_eventsource_js.js"))),
  ("ext:deno_http/00_serve.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_http_00_serve_ts.js"))),
  ("ext:deno_http/01_http.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_http_01_http_js.js"))),
  ("ext:deno_http/02_websocket.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_http_02_websocket_ts.js"))),
  ("ext:deno_kv/01_db.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_kv_01_db_ts.js"))),
  ("ext:deno_node/_events.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__events_mjs.js"))),
  ("ext:deno_node/_fs/_fs_common.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__fs__fs_common_ts.js"))),
  ("ext:deno_node/_fs/_fs_constants.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__fs__fs_constants_ts.js"))),
  ("ext:deno_node/_fs/_fs_cp.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__fs__fs_cp_ts.js"))),
  ("ext:deno_node/_fs/_fs_fstat.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__fs__fs_fstat_ts.js"))),
  ("ext:deno_node/_fs/_fs_lstat.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__fs__fs_lstat_ts.js"))),
  ("ext:deno_node/_fs/cp/cp.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__fs_cp_cp_ts.js"))),
  ("ext:deno_node/_fs/cp/cp_sync.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__fs_cp_cp_sync_ts.js"))),
  ("ext:deno_node/_next_tick.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__next_tick_ts.js"))),
  ("ext:deno_node/_process/exiting.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__process_exiting_ts.js"))),
  ("ext:deno_node/_process/process.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__process_process_ts.js"))),
  ("ext:deno_node/_repl_preview.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__repl_preview_js.js"))),
  ("ext:deno_node/_tls_common.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__tls_common_ts.js"))),
  ("ext:deno_node/_tls_wrap.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__tls_wrap_js.js"))),
  ("ext:deno_node/_util/_util_callbackify.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__util__util_callbackify_js.js"))),
  ("ext:deno_node/_util/os.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__util_os_ts.js"))),
  ("ext:deno_node/_utils.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__utils_ts.js"))),
  ("ext:deno_node/_zlib_binding.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__zlib_binding_mjs.js"))),
  ("ext:deno_node/assert.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_assert_ts.js"))),
  ("ext:deno_node/async_hooks.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_async_hooks_ts.js"))),
  ("ext:deno_node/child_process.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_child_process_ts.js"))),
  ("ext:deno_node/cluster.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_cluster_ts.js"))),
  ("ext:deno_node/console.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_console_ts.js"))),
  ("ext:deno_node/constants.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_constants_ts.js"))),
  ("ext:deno_node/crypto.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_crypto_ts.js"))),
  ("ext:deno_node/dgram.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_dgram_ts.js"))),
  ("ext:deno_node/diagnostics_channel.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_diagnostics_channel_js.js"))),
  ("ext:deno_node/dns.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_dns_ts.js"))),
  ("ext:deno_node/dns/promises.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_dns_promises_ts.js"))),
  ("ext:deno_node/domain.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_domain_ts.js"))),
  ("ext:deno_node/fs.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_fs_ts.js"))),
  ("ext:deno_node/fs/promises.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_fs_promises_ts.js"))),
  ("ext:deno_node/http.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_http_ts.js"))),
  ("ext:deno_node/http2.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_http2_ts.js"))),
  ("ext:deno_node/https.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_https_ts.js"))),
  ("ext:deno_node/inspector.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_inspector_js.js"))),
  ("ext:deno_node/inspector/promises.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_inspector_promises_js.js"))),
  ("ext:deno_node/inspector_network_bridge.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_inspector_network_bridge_js.js"))),
  ("ext:deno_node/internal/assert.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_assert_mjs.js"))),
  ("ext:deno_node/internal/assert/assertion_error.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_assert_assertion_error_js.js"))),
  ("ext:deno_node/internal/assert/calltracker.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_assert_calltracker_js.js"))),
  ("ext:deno_node/internal/assert/myers_diff.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_assert_myers_diff_js.js"))),
  ("ext:deno_node/internal/assert/utils.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_assert_utils_ts.js"))),
  ("ext:deno_node/internal/async_hooks.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_async_hooks_ts.js"))),
  ("ext:deno_node/internal/blocklist.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_blocklist_mjs.js"))),
  ("ext:deno_node/internal/buffer.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_buffer_mjs.js"))),
  ("ext:deno_node/internal/child_process.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_child_process_ts.js"))),
  ("ext:deno_node/internal/cli_table.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_cli_table_ts.js"))),
  ("ext:deno_node/internal/cluster/child.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_cluster_child_ts.js"))),
  ("ext:deno_node/internal/cluster/linkedlist.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_cluster_linkedlist_ts.js"))),
  ("ext:deno_node/internal/cluster/primary.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_cluster_primary_ts.js"))),
  ("ext:deno_node/internal/cluster/round_robin_handle.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_cluster_round_robin_handle_ts.js"))),
  ("ext:deno_node/internal/cluster/shared_handle.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_cluster_shared_handle_ts.js"))),
  ("ext:deno_node/internal/cluster/utils.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_cluster_utils_ts.js"))),
  ("ext:deno_node/internal/cluster/worker.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_cluster_worker_ts.js"))),
  ("ext:deno_node/internal/console/constructor.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_console_constructor_mjs.js"))),
  ("ext:deno_node/internal/constants.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_constants_ts.js"))),
  ("ext:deno_node/internal/crypto/_keys.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto__keys_ts.js"))),
  ("ext:deno_node/internal/crypto/_randomBytes.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto__randomBytes_ts.js"))),
  ("ext:deno_node/internal/crypto/_randomFill.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto__randomFill_mjs.js"))),
  ("ext:deno_node/internal/crypto/_randomInt.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto__randomInt_ts.js"))),
  ("ext:deno_node/internal/crypto/certificate.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto_certificate_ts.js"))),
  ("ext:deno_node/internal/crypto/cipher.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto_cipher_ts.js"))),
  ("ext:deno_node/internal/crypto/diffiehellman.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto_diffiehellman_ts.js"))),
  ("ext:deno_node/internal/crypto/hash.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto_hash_ts.js"))),
  ("ext:deno_node/internal/crypto/hkdf.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto_hkdf_ts.js"))),
  ("ext:deno_node/internal/crypto/keygen.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto_keygen_ts.js"))),
  ("ext:deno_node/internal/crypto/keys.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto_keys_ts.js"))),
  ("ext:deno_node/internal/crypto/pbkdf2.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto_pbkdf2_ts.js"))),
  ("ext:deno_node/internal/crypto/random.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto_random_ts.js"))),
  ("ext:deno_node/internal/crypto/scrypt.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto_scrypt_ts.js"))),
  ("ext:deno_node/internal/crypto/sig.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto_sig_ts.js"))),
  ("ext:deno_node/internal/crypto/util.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto_util_ts.js"))),
  ("ext:deno_node/internal/crypto/x509.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_crypto_x509_ts.js"))),
  ("ext:deno_node/internal/deps/undici/undici.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_deps_undici_undici_js.js"))),
  ("ext:deno_node/internal/dgram.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_dgram_ts.js"))),
  ("ext:deno_node/internal/dns/promises.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_dns_promises_ts.js"))),
  ("ext:deno_node/internal/dns/utils.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_dns_utils_ts.js"))),
  ("ext:deno_node/internal/dtrace.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_dtrace_ts.js"))),
  ("ext:deno_node/internal/error_codes.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_error_codes_ts.js"))),
  ("ext:deno_node/internal/errors.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_errors_ts.js"))),
  ("ext:deno_node/internal/errors/error_source.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_errors_error_source_ts.js"))),
  ("ext:deno_node/internal/events/abort_listener.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_events_abort_listener_mjs.js"))),
  ("ext:deno_node/internal/fs/sync_write_stream.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_fs_sync_write_stream_js.js"))),
  ("ext:deno_node/internal/hide_stack_frames.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_hide_stack_frames_ts.js"))),
  ("ext:deno_node/internal/http.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_http_ts.js"))),
  ("ext:deno_node/internal/http2/compat.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_http2_compat_js.js"))),
  ("ext:deno_node/internal/http2/constants.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_http2_constants_ts.js"))),
  ("ext:deno_node/internal/http2/core.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_http2_core_ts.js"))),
  ("ext:deno_node/internal/http2/util.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_http2_util_ts.js"))),
  ("ext:deno_node/internal/idna.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_idna_ts.js"))),
  ("ext:deno_node/internal/js_stream_socket.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_js_stream_socket_js.js"))),
  ("ext:deno_node/internal/mime.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_mime_ts.js"))),
  ("ext:deno_node/internal/net.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_net_ts.js"))),
  ("ext:deno_node/internal/normalize_encoding.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_normalize_encoding_ts.js"))),
  ("ext:deno_node/internal/options.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_options_ts.js"))),
  ("ext:deno_node/internal/primordials.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_primordials_mjs.js"))),
  ("ext:deno_node/internal/priority_queue.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_priority_queue_ts.js"))),
  ("ext:deno_node/internal/process/active_resources.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_process_active_resources_ts.js"))),
  ("ext:deno_node/internal/process/per_thread.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_process_per_thread_mjs.js"))),
  ("ext:deno_node/internal/process/report.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_process_report_ts.js"))),
  ("ext:deno_node/internal/process/warning.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_process_warning_ts.js"))),
  ("ext:deno_node/internal/querystring.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_querystring_ts.js"))),
  ("ext:deno_node/internal/readline/callbacks.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_readline_callbacks_mjs.js"))),
  ("ext:deno_node/internal/readline/emitKeypressEvents.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_readline_emitKeypressEvents_mjs.js"))),
  ("ext:deno_node/internal/readline/interface.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_readline_interface_mjs.js"))),
  ("ext:deno_node/internal/readline/promises.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_readline_promises_mjs.js"))),
  ("ext:deno_node/internal/readline/symbols.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_readline_symbols_mjs.js"))),
  ("ext:deno_node/internal/readline/utils.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_readline_utils_mjs.js"))),
  ("ext:deno_node/internal/socketaddress.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_socketaddress_js.js"))),
  ("ext:deno_node/internal/stream_base_commons.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_stream_base_commons_ts.js"))),
  ("ext:deno_node/internal/streams/add-abort-signal.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_add_abort_signal_js.js"))),
  ("ext:deno_node/internal/streams/destroy.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_destroy_js.js"))),
  ("ext:deno_node/internal/streams/duplex.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_duplex_js.js"))),
  ("ext:deno_node/internal/streams/end-of-stream.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_end_of_stream_js.js"))),
  ("ext:deno_node/internal/streams/legacy.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_legacy_js.js"))),
  ("ext:deno_node/internal/streams/passthrough.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_passthrough_js.js"))),
  ("ext:deno_node/internal/streams/readable.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_readable_js.js"))),
  ("ext:deno_node/internal/streams/state.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_state_js.js"))),
  ("ext:deno_node/internal/streams/transform.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_transform_js.js"))),
  ("ext:deno_node/internal/streams/utils.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_utils_js.js"))),
  ("ext:deno_node/internal/streams/writable.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_writable_js.js"))),
  ("ext:deno_node/internal/test/binding.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_test_binding_ts.js"))),
  ("ext:deno_node/internal/test/reporters.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_test_reporters_ts.js"))),
  ("ext:deno_node/internal/timers.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_timers_mjs.js"))),
  ("ext:deno_node/internal/tls_common.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_tls_common_js.js"))),
  ("ext:deno_node/internal/url.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_url_ts.js"))),
  ("ext:deno_node/internal/util.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_util_mjs.js"))),
  ("ext:deno_node/internal/util/colorize.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_util_colorize_mjs.js"))),
  ("ext:deno_node/internal/util/colors.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_util_colors_ts.js"))),
  ("ext:deno_node/internal/util/comparisons.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_util_comparisons_ts.js"))),
  ("ext:deno_node/internal/util/debuglog.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_util_debuglog_ts.js"))),
  ("ext:deno_node/internal/util/inspect.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_util_inspect_mjs.js"))),
  ("ext:deno_node/internal/util/parse_args/parse_args.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_util_parse_args_parse_args_js.js"))),
  ("ext:deno_node/internal/util/parse_args/utils.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_util_parse_args_utils_js.js"))),
  ("ext:deno_node/internal/util/types.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_util_types_ts.js"))),
  ("ext:deno_node/internal/validators.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_validators_mjs.js"))),
  ("ext:deno_node/internal/webstreams/adapters.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_webstreams_adapters_js.js"))),
  ("ext:deno_node/internal/webstreams/readablestream.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_webstreams_readablestream_js.js"))),
  ("ext:deno_node/internal/webstreams/util.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_webstreams_util_js.js"))),
  ("ext:deno_node/internal/worker/js_transferable.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_worker_js_transferable_js.js"))),
  ("ext:deno_node/internal_binding/_libuv_winerror.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding__libuv_winerror_ts.js"))),
  ("ext:deno_node/internal_binding/_listen.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding__listen_ts.js"))),
  ("ext:deno_node/internal_binding/_node.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding__node_ts.js"))),
  ("ext:deno_node/internal_binding/_timingSafeEqual.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding__timingSafeEqual_ts.js"))),
  ("ext:deno_node/internal_binding/_utils.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding__utils_ts.js"))),
  ("ext:deno_node/internal_binding/ares.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_ares_ts.js"))),
  ("ext:deno_node/internal_binding/async_wrap.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_async_wrap_ts.js"))),
  ("ext:deno_node/internal_binding/block_list.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_block_list_ts.js"))),
  ("ext:deno_node/internal_binding/buffer.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_buffer_ts.js"))),
  ("ext:deno_node/internal_binding/cares_wrap.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_cares_wrap_ts.js"))),
  ("ext:deno_node/internal_binding/constants.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_constants_ts.js"))),
  ("ext:deno_node/internal_binding/crypto.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_crypto_ts.js"))),
  ("ext:deno_node/internal_binding/handle_wrap.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_handle_wrap_ts.js"))),
  ("ext:deno_node/internal_binding/http2.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_http2_ts.js"))),
  ("ext:deno_node/internal_binding/http_parser.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_http_parser_ts.js"))),
  ("ext:deno_node/internal_binding/inspector.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_inspector_js.js"))),
  ("ext:deno_node/internal_binding/node_file.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_node_file_ts.js"))),
  ("ext:deno_node/internal_binding/node_options.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_node_options_ts.js"))),
  ("ext:deno_node/internal_binding/pipe_wrap.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_pipe_wrap_ts.js"))),
  ("ext:deno_node/internal_binding/stream_wrap.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_stream_wrap_ts.js"))),
  ("ext:deno_node/internal_binding/string_decoder.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_string_decoder_ts.js"))),
  ("ext:deno_node/internal_binding/symbols.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_symbols_ts.js"))),
  ("ext:deno_node/internal_binding/tcp_wrap.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_tcp_wrap_ts.js"))),
  ("ext:deno_node/internal_binding/tls_wrap.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_tls_wrap_ts.js"))),
  ("ext:deno_node/internal_binding/tty_wrap.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_tty_wrap_ts.js"))),
  ("ext:deno_node/internal_binding/types.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_types_ts.js"))),
  ("ext:deno_node/internal_binding/udp_wrap.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_udp_wrap_ts.js"))),
  ("ext:deno_node/internal_binding/util.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_util_ts.js"))),
  ("ext:deno_node/internal_binding/uv.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_uv_ts.js"))),
  ("ext:deno_node/net.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_net_ts.js"))),
  ("ext:deno_node/os.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_os_ts.js"))),
  ("ext:deno_node/path/_constants.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_path__constants_ts.js"))),
  ("ext:deno_node/path/_interface.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_path__interface_ts.js"))),
  ("ext:deno_node/path/_posix.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_path__posix_ts.js"))),
  ("ext:deno_node/path/_util.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_path__util_ts.js"))),
  ("ext:deno_node/path/_win32.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_path__win32_ts.js"))),
  ("ext:deno_node/path/common.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_path_common_ts.js"))),
  ("ext:deno_node/path/mod.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_path_mod_ts.js"))),
  ("ext:deno_node/path/separator.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_path_separator_ts.js"))),
  ("ext:deno_node/perf_hooks.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_perf_hooks_js.js"))),
  ("ext:deno_node/punycode.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_punycode_ts.js"))),
  ("ext:deno_node/querystring.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_querystring_js.js"))),
  ("ext:deno_node/sqlite.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_sqlite_ts.js"))),
  ("ext:deno_node/stream/consumers.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_stream_consumers_js.js"))),
  ("ext:deno_node/stream/web.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_stream_web_js.js"))),
  ("ext:deno_node/string_decoder.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_string_decoder_ts.js"))),
  ("ext:deno_node/test/reporters.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_test_reporters_ts.js"))),
  ("ext:deno_node/testing.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_testing_ts.js"))),
  ("ext:deno_node/timers.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_timers_ts.js"))),
  ("ext:deno_node/timers/promises.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_timers_promises_ts.js"))),
  ("ext:deno_node/tls.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_tls_ts.js"))),
  ("ext:deno_node/trace_events.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_trace_events_ts.js"))),
  ("ext:deno_node/tty.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_tty_js.js"))),
  ("ext:deno_node/url.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_url_ts.js"))),
  ("ext:deno_node/util.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_util_ts.js"))),
  ("ext:deno_node/v8.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_v8_ts.js"))),
  ("ext:deno_node/vm.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_vm_js.js"))),
  ("ext:deno_node/wasi.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_wasi_ts.js"))),
  ("ext:deno_node/worker_threads.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_worker_threads_ts.js"))),
  ("ext:deno_node/zlib.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_zlib_js.js"))),
  ("ext:deno_process/40_process.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_process_40_process_js.js"))),
  ("ext:deno_telemetry/telemetry.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_telemetry_telemetry_ts.js"))),
  ("ext:deno_telemetry/util.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_telemetry_util_ts.js"))),
  ("ext:deno_web/01_broadcast_channel.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_web_01_broadcast_channel_js.js"))),
  ("ext:deno_web/01_mimesniff.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_web_01_mimesniff_js.js"))),
  ("ext:deno_web/01_urlpattern.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_web_01_urlpattern_js.js"))),
  ("ext:deno_web/02_structured_clone.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_web_02_structured_clone_js.js"))),
  ("ext:deno_web/05_base64.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_web_05_base64_js.js"))),
  ("ext:deno_web/06_streams.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_web_06_streams_js.js"))),
  ("ext:deno_web/10_filereader.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_web_10_filereader_js.js"))),
  ("ext:deno_web/14_compression.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_web_14_compression_js.js"))),
  ("ext:deno_web/16_image_data.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_web_16_image_data_js.js"))),
  ("ext:deno_web/17_geometry.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_web_17_geometry_js.js"))),
  ("ext:deno_web/18_css_stylesheet.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_web_18_css_stylesheet_js.js"))),
  ("ext:deno_webstorage/01_webstorage.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_webstorage_01_webstorage_js.js"))),
  ("ext:runtime/41_prompt.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_runtime_41_prompt_js.js"))),
];

pub static RESIDUAL_LAZY_ESM: &[(&str, &str)] = &[
  ("ext:deno_canvas/01_canvas.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_canvas_01_canvas_js.js"))),
  ("ext:deno_image/01_image.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_image_01_image_js.js"))),
  ("ext:deno_net/03_quic.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_net_03_quic_js.js"))),
  ("ext:deno_node/_fs/_fs_copy.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__fs__fs_copy_ts.js"))),
  ("ext:deno_node/_fs/_fs_dir.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__fs__fs_dir_ts.js"))),
  ("ext:deno_node/_fs/_fs_exists.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__fs__fs_exists_ts.js"))),
  ("ext:deno_node/_fs/_fs_glob.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__fs__fs_glob_ts.js"))),
  ("ext:deno_node/_fs/_fs_lutimes.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__fs__fs_lutimes_ts.js"))),
  ("ext:deno_node/_fs/_fs_read.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__fs__fs_read_ts.js"))),
  ("ext:deno_node/_fs/_fs_readdir.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__fs__fs_readdir_ts.js"))),
  ("ext:deno_node/_process/streams.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__process_streams_mjs.js"))),
  ("ext:deno_node/_readline.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node__readline_mjs.js"))),
  ("ext:deno_node/deps/minimatch.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_deps_minimatch_js.js"))),
  ("ext:deno_node/internal/event_target.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_event_target_mjs.js"))),
  ("ext:deno_node/internal/fs/handle.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_fs_handle_ts.js"))),
  ("ext:deno_node/internal/fs/promises.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_fs_promises_ts.js"))),
  ("ext:deno_node/internal/fs/stat_utils.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_fs_stat_utils_ts.js"))),
  ("ext:deno_node/internal/fs/streams.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_fs_streams_mjs.js"))),
  ("ext:deno_node/internal/fs/utils.mjs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_fs_utils_mjs.js"))),
  ("ext:deno_node/internal/http/address_override.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_http_address_override_js.js"))),
  ("ext:deno_node/internal/repl.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_repl_ts.js"))),
  ("ext:deno_node/internal/streams/compose.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_compose_js.js"))),
  ("ext:deno_node/internal/streams/duplexify.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_duplexify_js.js"))),
  ("ext:deno_node/internal/streams/duplexpair.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_duplexpair_js.js"))),
  ("ext:deno_node/internal/streams/fast-utf8-stream.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_fast_utf8_stream_js.js"))),
  ("ext:deno_node/internal/streams/from.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_from_js.js"))),
  ("ext:deno_node/internal/streams/lazy_transform.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_lazy_transform_js.js"))),
  ("ext:deno_node/internal/streams/operators.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_operators_js.js"))),
  ("ext:deno_node/internal/streams/pipeline.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_streams_pipeline_js.js"))),
  ("ext:deno_node/internal/tty.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_tty_js.js"))),
  ("ext:deno_node/internal_binding/mod.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_internal_binding_mod_ts.js"))),
  ("ext:deno_node/readline/promises.ts", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_node_readline_promises_ts.js"))),
  ("ext:deno_web/locks.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_web_locks_js.js"))),
  ("ext:deno_web/webtransport.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_web_webtransport_js.js"))),
  ("ext:deno_webgpu/01_webgpu.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_webgpu_01_webgpu_js.js"))),
  ("ext:deno_websocket/01_websocket.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_websocket_01_websocket_js.js"))),
  ("ext:deno_websocket/02_websocketstream.js", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/ext_deno_websocket_02_websocketstream_js.js"))),
  ("node:_http_agent", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node__http_agent.js"))),
  ("node:_http_client", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node__http_client.js"))),
  ("node:_http_common", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node__http_common.js"))),
  ("node:_http_incoming", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node__http_incoming.js"))),
  ("node:_http_outgoing", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node__http_outgoing.js"))),
  ("node:_http_proxy", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node__http_proxy.js"))),
  ("node:_http_server", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node__http_server.js"))),
  ("node:_stream_duplex", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node__stream_duplex.js"))),
  ("node:_stream_passthrough", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node__stream_passthrough.js"))),
  ("node:_stream_readable", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node__stream_readable.js"))),
  ("node:_stream_transform", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node__stream_transform.js"))),
  ("node:_stream_writable", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node__stream_writable.js"))),
  ("node:_tls_common", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node__tls_common.js"))),
  ("node:_tls_wrap", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node__tls_wrap.js"))),
  ("node:assert", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_assert.js"))),
  ("node:assert/strict", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_assert_strict.js"))),
  ("node:async_hooks", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_async_hooks.js"))),
  ("node:buffer", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_buffer.js"))),
  ("node:child_process", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_child_process.js"))),
  ("node:cluster", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_cluster.js"))),
  ("node:console", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_console.js"))),
  ("node:constants", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_constants.js"))),
  ("node:crypto", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_crypto.js"))),
  ("node:dgram", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_dgram.js"))),
  ("node:diagnostics_channel", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_diagnostics_channel.js"))),
  ("node:dns", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_dns.js"))),
  ("node:dns/promises", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_dns_promises.js"))),
  ("node:domain", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_domain.js"))),
  ("node:events", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_events.js"))),
  ("node:fs", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_fs.js"))),
  ("node:fs/promises", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_fs_promises.js"))),
  ("node:http", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_http.js"))),
  ("node:http2", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_http2.js"))),
  ("node:https", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_https.js"))),
  ("node:inspector", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_inspector.js"))),
  ("node:inspector/promises", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_inspector_promises.js"))),
  ("node:module", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_module.js"))),
  ("node:net", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_net.js"))),
  ("node:os", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_os.js"))),
  ("node:path", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_path.js"))),
  ("node:path/posix", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_path_posix.js"))),
  ("node:path/win32", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_path_win32.js"))),
  ("node:perf_hooks", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_perf_hooks.js"))),
  ("node:process", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_process.js"))),
  ("node:punycode", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_punycode.js"))),
  ("node:querystring", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_querystring.js"))),
  ("node:readline", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_readline.js"))),
  ("node:readline/promises", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_readline_promises.js"))),
  ("node:repl", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_repl.js"))),
  ("node:sqlite", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_sqlite.js"))),
  ("node:stream", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_stream.js"))),
  ("node:stream/consumers", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_stream_consumers.js"))),
  ("node:stream/promises", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_stream_promises.js"))),
  ("node:stream/web", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_stream_web.js"))),
  ("node:string_decoder", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_string_decoder.js"))),
  ("node:sys", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_sys.js"))),
  ("node:test", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_test.js"))),
  ("node:test/reporters", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_test_reporters.js"))),
  ("node:timers", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_timers.js"))),
  ("node:timers/promises", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_timers_promises.js"))),
  ("node:tls", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_tls.js"))),
  ("node:trace_events", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_trace_events.js"))),
  ("node:tty", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_tty.js"))),
  ("node:util/types", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_util_types.js"))),
  ("node:v8", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_v8.js"))),
  ("node:vm", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_vm.js"))),
  ("node:wasi", include_str!(concat!(env!("OUT_DIR"), "/residual_sources/node_wasi.js"))),
];

