// Generated by `generate::isle_macro`, /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/lib.rs:22
#[macro_export] // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate.rs:38
macro_rules! isle_assembler_methods {
  () => {
    fn x64_andb_i(
      &mut self,
      imm8: &cranelift_assembler_x64::Imm8,
    ) -> cranelift_assembler_x64::Gpr<PairedGpr> {
      let inst =
        cranelift_assembler_x64::inst::andb_i::new(imm8.clone()).into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      todo!()
    }
    fn x64_andw_i(
      &mut self,
      imm16: &cranelift_assembler_x64::Imm16,
    ) -> cranelift_assembler_x64::Gpr<PairedGpr> {
      let inst =
        cranelift_assembler_x64::inst::andw_i::new(imm16.clone()).into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      todo!()
    }
    fn x64_andl_i(
      &mut self,
      imm32: &cranelift_assembler_x64::Imm32,
    ) -> cranelift_assembler_x64::Gpr<PairedGpr> {
      let inst =
        cranelift_assembler_x64::inst::andl_i::new(imm32.clone()).into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      todo!()
    }
    fn x64_andq_i_sxlq(
      &mut self,
      imm32: &cranelift_assembler_x64::Imm32,
    ) -> cranelift_assembler_x64::Gpr<PairedGpr> {
      let inst =
        cranelift_assembler_x64::inst::andq_i_sxlq::new(imm32.clone()).into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      todo!()
    }
    fn x64_andb_mi(
      &mut self,
      rm8: &cranelift_assembler_x64::GprMem<PairedGpr, Gpr>,
      imm8: &cranelift_assembler_x64::Imm8,
    ) -> cranelift_assembler_x64::GprMem<PairedGpr, Gpr> {
      let inst =
        cranelift_assembler_x64::inst::andb_mi::new(rm8.clone(), imm8.clone())
          .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      rm8.clone()
    }
    fn x64_andw_mi(
      &mut self,
      rm16: &cranelift_assembler_x64::GprMem<PairedGpr, Gpr>,
      imm16: &cranelift_assembler_x64::Imm16,
    ) -> cranelift_assembler_x64::GprMem<PairedGpr, Gpr> {
      let inst = cranelift_assembler_x64::inst::andw_mi::new(
        rm16.clone(),
        imm16.clone(),
      )
      .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      rm16.clone()
    }
    fn x64_andl_mi(
      &mut self,
      rm32: &cranelift_assembler_x64::GprMem<PairedGpr, Gpr>,
      imm32: &cranelift_assembler_x64::Imm32,
    ) -> cranelift_assembler_x64::GprMem<PairedGpr, Gpr> {
      let inst = cranelift_assembler_x64::inst::andl_mi::new(
        rm32.clone(),
        imm32.clone(),
      )
      .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      rm32.clone()
    }
    fn x64_andq_mi_sxlq(
      &mut self,
      rm64: &cranelift_assembler_x64::GprMem<PairedGpr, Gpr>,
      imm32: &cranelift_assembler_x64::Imm32,
    ) -> cranelift_assembler_x64::GprMem<PairedGpr, Gpr> {
      let inst = cranelift_assembler_x64::inst::andq_mi_sxlq::new(
        rm64.clone(),
        imm32.clone(),
      )
      .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      rm64.clone()
    }
    fn x64_andl_mi_sxbl(
      &mut self,
      rm32: &cranelift_assembler_x64::GprMem<PairedGpr, Gpr>,
      imm8: &cranelift_assembler_x64::Imm8,
    ) -> cranelift_assembler_x64::GprMem<PairedGpr, Gpr> {
      let inst = cranelift_assembler_x64::inst::andl_mi_sxbl::new(
        rm32.clone(),
        imm8.clone(),
      )
      .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      rm32.clone()
    }
    fn x64_andq_mi_sxbq(
      &mut self,
      rm64: &cranelift_assembler_x64::GprMem<PairedGpr, Gpr>,
      imm8: &cranelift_assembler_x64::Imm8,
    ) -> cranelift_assembler_x64::GprMem<PairedGpr, Gpr> {
      let inst = cranelift_assembler_x64::inst::andq_mi_sxbq::new(
        rm64.clone(),
        imm8.clone(),
      )
      .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      rm64.clone()
    }
    fn x64_andb_mr(
      &mut self,
      rm8: &cranelift_assembler_x64::GprMem<PairedGpr, Gpr>,
      r8: &cranelift_assembler_x64::Gpr<Gpr>,
    ) -> cranelift_assembler_x64::GprMem<PairedGpr, Gpr> {
      let inst =
        cranelift_assembler_x64::inst::andb_mr::new(rm8.clone(), r8.clone())
          .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      rm8.clone()
    }
    fn x64_andb_mr_sxbq(
      &mut self,
      rm8: &cranelift_assembler_x64::GprMem<PairedGpr, Gpr>,
      r8: &cranelift_assembler_x64::Gpr<Gpr>,
    ) -> cranelift_assembler_x64::GprMem<PairedGpr, Gpr> {
      let inst = cranelift_assembler_x64::inst::andb_mr_sxbq::new(
        rm8.clone(),
        r8.clone(),
      )
      .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      rm8.clone()
    }
    fn x64_andw_mr(
      &mut self,
      rm16: &cranelift_assembler_x64::GprMem<PairedGpr, Gpr>,
      r16: &cranelift_assembler_x64::Gpr<Gpr>,
    ) -> cranelift_assembler_x64::GprMem<PairedGpr, Gpr> {
      let inst =
        cranelift_assembler_x64::inst::andw_mr::new(rm16.clone(), r16.clone())
          .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      rm16.clone()
    }
    fn x64_andl_mr(
      &mut self,
      rm32: &cranelift_assembler_x64::GprMem<PairedGpr, Gpr>,
      r32: &cranelift_assembler_x64::Gpr<Gpr>,
    ) -> cranelift_assembler_x64::GprMem<PairedGpr, Gpr> {
      let inst =
        cranelift_assembler_x64::inst::andl_mr::new(rm32.clone(), r32.clone())
          .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      rm32.clone()
    }
    fn x64_andq_mr(
      &mut self,
      rm64: &cranelift_assembler_x64::GprMem<PairedGpr, Gpr>,
      r64: &cranelift_assembler_x64::Gpr<Gpr>,
    ) -> cranelift_assembler_x64::GprMem<PairedGpr, Gpr> {
      let inst =
        cranelift_assembler_x64::inst::andq_mr::new(rm64.clone(), r64.clone())
          .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      rm64.clone()
    }
    fn x64_andb_rm(
      &mut self,
      r8: &cranelift_assembler_x64::Gpr<PairedGpr>,
      rm8: &cranelift_assembler_x64::GprMem<Gpr, Gpr>,
    ) -> cranelift_assembler_x64::Gpr<PairedGpr> {
      let inst =
        cranelift_assembler_x64::inst::andb_rm::new(r8.clone(), rm8.clone())
          .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      r8.clone()
    }
    fn x64_andb_rm_sxbq(
      &mut self,
      r8: &cranelift_assembler_x64::Gpr<PairedGpr>,
      rm8: &cranelift_assembler_x64::GprMem<Gpr, Gpr>,
    ) -> cranelift_assembler_x64::Gpr<PairedGpr> {
      let inst = cranelift_assembler_x64::inst::andb_rm_sxbq::new(
        r8.clone(),
        rm8.clone(),
      )
      .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      r8.clone()
    }
    fn x64_andw_rm(
      &mut self,
      r16: &cranelift_assembler_x64::Gpr<PairedGpr>,
      rm16: &cranelift_assembler_x64::GprMem<Gpr, Gpr>,
    ) -> cranelift_assembler_x64::Gpr<PairedGpr> {
      let inst =
        cranelift_assembler_x64::inst::andw_rm::new(r16.clone(), rm16.clone())
          .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      r16.clone()
    }
    fn x64_andl_rm(
      &mut self,
      r32: &cranelift_assembler_x64::Gpr<PairedGpr>,
      rm32: &cranelift_assembler_x64::GprMem<Gpr, Gpr>,
    ) -> cranelift_assembler_x64::Gpr<PairedGpr> {
      let inst =
        cranelift_assembler_x64::inst::andl_rm::new(r32.clone(), rm32.clone())
          .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      r32.clone()
    }
    fn x64_andq_rm(
      &mut self,
      r64: &cranelift_assembler_x64::Gpr<PairedGpr>,
      rm64: &cranelift_assembler_x64::GprMem<Gpr, Gpr>,
    ) -> cranelift_assembler_x64::Gpr<PairedGpr> {
      let inst =
        cranelift_assembler_x64::inst::andq_rm::new(r64.clone(), rm64.clone())
          .into(); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:260
      self.lower_ctx.emit(MInst::External { inst }); // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate/inst.rs:261
      r64.clone()
    }
  }; // /build/.cargo/registry/src/index.crates.io-1949cf8c6b5b557f/cranelift-assembler-x64-meta-0.117.2/src/generate.rs:47
}
